from retrieval.retriever import Retriever
from elasticsearch_dsl import Search, Q
from elasticsearch_dsl.connections import connections
from elasticsearch_dsl import Document, Text, connections, Integer, Float, Keyword
import pandas as pd
import logging

logging.basicConfig(format='%(levelname)s :: %(asctime)s :: %(message)s', level=logging.WARNING)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Object(Document):
    cls = Text(fields={'raw': Keyword()})
    score = Float()
    dataset_id = Text(fields={'raw': Keyword()})
    content = Text()
    area = Integer()
    pdf_name = Text(fields={'raw': Keyword()})
    page_num = Integer()

    class Index:
        name = 'object'
        settings = {
            'number_of_shards': 1,
            'number_of_replicas': 0
        }


class FullDocument(Document):
    dataset_id = Text(fields={'raw': Keyword()})
    content = Text()
    name = Text(fields={'raw': Keyword()})

    class Index:
        name = 'fulldocument'
        settings = {
            'number_of_shards': 1,
            'number_of_replicas': 0
        }


class ElasticRetriever(Retriever):
    def __init__(self, hosts=['localhost']):
        self.hosts = hosts

    def search(self, query):
        connections.create_connection(hosts=self.hosts)
        q = Q('match', content=query)
        s = Search(index='fulldocument').query(q)[:30]
        response = s.execute()
        contexts = []
        for result in response:
            q = Q('match', pdf_name__raw=result['name'])
            s = Search(index='object').query(q)
            for context in s.scan():
                contexts.append({'id': context.meta.id, 'pdf_name': context['pdf_name'], 'content': context['content']})
        return contexts

    def build_index(self, document_parquet, section_parquet):
        logger.info('Building elastic index')
        connections.create_connection(hosts=self.hosts)
        Object.init()
        FullDocument.init()
        # This is a parquet file to load from
        df = pd.read_parquet(document_parquet)
        for ind, row in df.iterrows():
            FullDocument(name=row['pdf_name'], dataset_id='none', content=row['content']).save()
        logger.info('Done building document index')
        df = pd.read_parquet(section_parquet)
        for ind, row in df.iterrows():
            Object(cls='Section',
                   dataset_id='none', #row['dataset_id'],
                   content=row['content'],
                   pdf_name=row['pdf_name'],
                   ).save()
        logger.info('Done building object index')

    def delete(self, dataset_id):
        connections.create_connection(hosts=self.hosts)
        s = Search(index='fulldocument')
        q = Q()
        q = q & Q('match', dataset_id=dataset_id)
        result = s.query(q).delete()
        logger.info(result)
        s = Search(index='object')
        q = Q()
        q = q & Q('match', dataset_id=dataset_id)
        result = s.query(q).delete()
        logger.info(result)

    def rerank(self, query, contexts):
        raise NotImplementedError('ElasticRetriever does not rerank results')



from retrieval.retriever import Retriever
from retrieval.bert_reranker.inference import Inference
from transformers import BertForSequenceClassification
from distributed.diagnostics.plugin import WorkerPlugin
import os
import click
from dask.distributed import Client
from dask.distributed import get_worker
import logging
logging.basicConfig(format='%(levelname)s :: %(asctime)s :: %(message)s', level=logging.WARNING)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class BertRerankingRetriever(Retriever):
    def __init__(self, client):
        if type(client) == str:
            logger.info('Input is a str, inferring to be scheduler address. Initializing client')
            self.addr = client
            self.client = Client(client)#, serializers=['msgpack', 'dask'], deserializers=['msgpack', 'dask'])
            logger.info(self.client)
        else:
            self.client = client
            self.addr = None
        
    def __del__(self):
        if self.addr is not None:
            self.client.close()

    def search(self, query):
        raise NotImplementedError('Reranking does not search')

    def rerank(self, query, contexts):
        result = self.client.submit(BertRerankingRetriever._rerank, query, contexts, resources={'retrieval': 1})
        result = result.result()
        return result

    @classmethod
    def _rerank(cls, query, contexts):
        contexts = [c for c in contexts if c['content'] != '']
        worker = get_worker()
        dp = None
        for plg in worker.plugins:
            if 'bert_reranking_retriever' in plg:
                dp = worker.plugins[plg]
                break
        if dp is None:
            raise Exception('No reranking plugin registered')
        model = dp.model
        print(contexts)
        result = model.infer(query, contexts)
        print(result)
        #contexts, scores = zip(*result)
        #result = {'contexts': contexts, 'scores': scores}
        return result

    def build_index(self, path):
        raise NotImplementedError('Reranking does not require building an index')

    def delete(self, path):
        raise NotImplementedError('Reranking does not require building an index')


class BertRerankingPlugin(WorkerPlugin):
    def __init__(self):
        model_path = os.environ.get("RERANKING_MODEL_PATH")
        device = os.environ.get("RERANKING_DEVICE")
        bsz = int(os.environ.get("RERANKING_BATCH_SIZE"))
        num_workers = int(os.environ.get("RERANKING_NUM_WORKERS"))
        model = BertForSequenceClassification.from_pretrained(model_path)
        self.model = Inference(model, model_path, bsz, device, num_workers)

@click.command()
def dask_setup(worker):
    worker._pending_plugins = [BertRerankingPlugin()]




import torch
from torch.nn.utils.rnn import pad_sequence
from typing import Dict

Encoding = Dict[str, torch.Tensor]
CUDA = torch.cuda.is_available()

def remove_padding(xs: Encoding, pad_token) -> Encoding:
    mask = xs["input_ids"] != pad_token
    if mask.ndim == 1 or mask.shape[0] == 1:
        return {k: v[mask] for k, v in xs.items()}
    else:
        max_len = mask.sum(-1).max()
        return {k: v[:, :max_len] for k, v in xs.items()}

def collate(batch, padding_value=0):
    """simplified version of the pytorch default_collate_fn that deals with padding"""

    elem = batch[0]

    if isinstance(elem, torch.Tensor):
        return pad_sequence(batch, True, padding_value=padding_value)
    elif isinstance(elem, dict):
        return {k: collate([d[k] for d in batch], padding_value=padding_value) for k in elem}
    elif isinstance(elem, tuple):
        if type(elem) is not tuple:
            # elem is a NamedTuple
            cls = elem.__class__
            return cls(*(collate(samples, padding_value) for samples in zip(*batch)))
        else:
            return tuple(collate(samples, padding_value) for samples in zip(*batch))
    elif isinstance(elem, int):
        return torch.tensor(batch, dtype=torch.long).unsqueeze(-1)
    elif isinstance(elem, float):
        return torch.tensor(batch, dtype=torch.float32).unsqueeze(-1)
    else:
        raise TypeError(type(elem))

def cudafy(thing):
    if not CUDA:
        return thing
    if isinstance(thing, torch.Tensor):
        return thing.cuda()
    elif isinstance(thing, dict):
        return {k: cudafy(v) for k, v in thing.items()}
    elif isinstance(thing, tuple):
        cls = thing.__class__  # so this works on named tuples
        return cls(*(cudafy(t) for t in thing))
    elif isinstance(thing, list):
        return [cudafy(t) for t in thing]
    else:
        raise TypeError(type(thing))




import json
import shutil
import argparse
import os
import subprocess
from schema import ObjectContext, Pdf
from sqlalchemy.sql import text
import pickle
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
import logging
logging.basicConfig(format='%(levelname)s :: %(asctime)s :: %(message)s', level=logging.DEBUG)
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


def convert(did, session, conn):
    for ctype in ["Section", "FigureContext", "TableContext", "EquationContext", "Combined"]:
        logging.info(f"Indexing {ctype}")
        dataset_pth = os.path.join('/index_dir', did, ctype)
        if os.path.exists(dataset_pth):
            logger.warning(f'Found the dataset directory already created {dataset_pth}. Deleting contents to repopulate')
            shutil.rmtree(dataset_pth)
        os.makedirs(dataset_pth)
        os.makedirs(os.path.join(dataset_pth, 'documents_jsonl'))
        os.makedirs(os.path.join(dataset_pth, 'contexts_jsonl'))

        id_map = {}
        id_context_map = {}
        i = 0
        # Full contexts
        if ctype == "Combined" :
            r = session.query(ObjectContext.id, ObjectContext.content, ObjectContext.pdf_id).filter(ObjectContext.id, ObjectContext.pdf_id == Pdf.id).filter(Pdf.dataset_id == did)
        else:
            r = session.query(ObjectContext.id, ObjectContext.content, ObjectContext.pdf_id).filter(ObjectContext.id, ObjectContext.pdf_id == Pdf.id).filter(Pdf.dataset_id == did).filter(ObjectContext.cls == ctype)
        with open(os.path.join(dataset_pth, 'contexts_jsonl', 'full_contexts.jsonl'), 'w') as wf, open(os.path.join(dataset_pth, 'raw.txt'), 'w') as rawf, open(os.path.join(dataset_pth, 'split_contexts.jsonl'), 'w') as splf:
            for id, content, pdfid in r:
                id_map[id] = pdfid

                example = {'id': i, 'contents': content}
                id_context_map[i] = str(id)
                i += 1
                estr = json.dumps(example)
                wf.write(f'{estr}\n')
                rawf.write(f'{content}\n')
                # Sliding window, stride 64, window 256 (subword splitting downstream, assumption: avg 2 subwords / word)
                spl_content = content.split(' ')
                start = 0
                end = 256
                while end < len(spl_content):
                    example = {'id': i, 'contents': ' '.join(spl_content[start:end])}
                    id_context_map[i] = str(id)
                    i += 1
                    estr = json.dumps(example)
                    splf.write(f'{estr}\n')
                    end += 64
                    start += 64
                final_example = {'id': i, 'contents': ' '.join(spl_content[start:])}
                id_context_map[i] = str(id)
                i += 1
                estr = json.dumps(final_example)
                splf.write(f'{estr}\n')
        with open(os.path.join(dataset_pth, 'context_to_doc.pkl'), 'wb') as wp1, open(os.path.join(dataset_pth, 'id_to_context.pkl'), 'wb') as wp2:
            pickle.dump(id_map, wp1)
            pickle.dump(id_context_map, wp2)

        # Just going to write some SQL here
        q = text('''SELECT pdfs.id, GROUP_CONCAT(object_contexts.content SEPARATOR '\n')
                    FROM object_contexts, pdfs
                    WHERE object_contexts.pdf_id = pdfs.id
                    GROUP BY pdfs.id
                    ''')
        r2 = conn.execute(q)
        i = 0
        id_to_pdf = {}
        with open(os.path.join(dataset_pth, 'documents_jsonl', 'documents.jsonl'), 'w') as wf:
            for pdf_id, content in r2:
                example = {'id': i, 'contents': content}
                id_to_pdf[i] = str(pdf_id)
                i += 1
                estr = json.dumps(example)
                wf.write(f'{estr}\n')
        with open(os.path.join(dataset_pth, 'id_to_pdf.pkl'), 'wb') as wp1:
            pickle.dump(id_to_pdf, wp1)
        logger.info('done writing files')
        logger.info('Starting indexing')
        subprocess.run(['/bin/bash', '-c', f'./index.sh {os.path.join(dataset_pth, "documents_jsonl")}'])
        subprocess.run(['/bin/bash', '-c', f'./index.sh {os.path.join(dataset_pth, "contexts_jsonl")}'])



if __name__ == '__main__':
    engine = create_engine(f'mysql://{os.environ["MYSQL_USER"]}:{os.environ["MYSQL_PASSWORD"]}@mysql-router:6446/cosmos', pool_pre_ping=True)
    did = os.environ["DATASET_ID"]
    conn = engine.connect()
    Session = sessionmaker()
    Session.configure(bind=engine)
    session = Session()
    convert(did, session, conn)






from retrieval.retriever import Retriever
from retrieval.bert_reranker.bert_reranking_retriever import BertRerankingRetriever
from retrieval.elastic_retriever import ElasticRetriever
import logging
logging.basicConfig(format='%(levelname)s :: %(asctime)s :: %(message)s', level=logging.WARNING)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
import os


class ElasticRerankingRetriever(Retriever):
    def __init__(self, client, hosts=[os.environ["ELASTIC_ADDRESS"]]):
        self.elastic_retriever = ElasticRetriever(hosts)
        self.reranker = BertRerankingRetriever(client)

    def search(self, query):
        contexts = self.elastic_retriever.search(query)
        return self.rerank(query, contexts)

    def rerank(self, query, contexts):
        return self.reranker.rerank(query, contexts)

    def build_index(self, document_parquet, section_parquet):
        self.elastic_retriever.build_index(document_parquet, section_parquet)

    def delete(self, dataset_id):
        self.elastic_retriever.delete(dataset_id)




import json
import argparse
import os 

#def convert_dataset(args):
#    print("Converting dataset to Anserini jsonl file")
#    file_index = 0
#    i = 0
#    # id_map = {} # TODO: Use this if you want a mapping between id in anserini indexing to dataset id
#    #TODO: Load the dataset using args.dataset_path
#    #for each section in the dataset:
#    #    content = "" # TODO: Read the section content from dataset
#    #    # id_map[i] = <section_id_from_dataset> # TODO: Use this if index_table required
#    #    if i % args.max_docs_per_file == 0:
#    #        if i > 0:
#    #            output_jsonl_file.close()
#    #        output_path = os.path.join(args.output_folder, 'docs{:02d}.json'.format(file_index))
#    #        output_jsonl_file = open(output_path, 'w', encoding='utf-8', newline='\n')
#    #        file_index += 1
#    #    output_dict = {'id': str(i), 'contents': content}
#    #    output_jsonl_file.write(json.dumps(output_dict) + '\n')
#
#    #    if i % 100000 == 0:
#    #        print('Converted {} docs in {} files'.format(i, file_index))
#    #    i += 1
#    #output_jsonl_file.close()
#
#    # TODO: Use the following if id mapping required for any reason
#    # with open(args.output_folder+'/id_map.json', 'w') as fp:
#    #     json.dump(id_map, fp)
#
#if __name__=='__main__':
#
#    parser = argparse.ArgumentParser(description='''Converts processedx dataset to Anserini jsonl files.''')
#    parser.add_argument('--dataset_path', required=True, help='path to dataset')
#    parser.add_argument('--output_folder', required=True, help='output file')
#    parser.add_argument('--max_docs_per_file', default=1000000, type=int, help='maximum number of documents in each jsonl file.')
#
#    args = parser.parse_args()
#
#    if not os.path.exists(args.output_folder):
#        os.makedirs(args.output_folder)
#
#    convert_dataset(args)
#    print('Done!')
#    
#



from retrieval.bert_reranker.inference_dataset import InferenceDataset
from transformers import BertTokenizerFast
from retrieval.utils import cudafy
from torch.utils.data import DataLoader
import glob
import os
import numpy as np
import logging
logging.basicConfig(format='%(levelname)s :: %(asctime)s :: %(message)s', level=logging.WARNING)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class Inference:
    def __init__(self, model, model_pth, bsz, device, num_workers):
        self.model = model.to(device)
        self.model.eval()
        self.bsz = bsz
        self.num_workers = num_workers
        self.tokenizer = BertTokenizerFast.from_pretrained(model_pth)
        logger.error(self.tokenizer.pad_token_id)
                                          

    def infer(self, query, contexts):
        dataset = InferenceDataset(query, contexts, self.tokenizer)
        dataloader = DataLoader(dataset, 
                                batch_size=self.bsz, 
                                shuffle=False, 
                                collate_fn=InferenceDataset.collate,)
        full_scores = []
        full_ids = []
        full_docnames = []
        for batch in dataloader:
            xs, ids, docnames = batch
            xs = cudafy(xs)
            logits = self.model(**xs)[0]
            scores = logits[:, 1]
            scores = scores.squeeze()
            scores = list(scores.detach().cpu().numpy().flatten())
            full_scores.extend(scores)
            full_ids.extend(ids)
            full_docnames.extend(docnames)
        cs = zip(contexts, full_scores, full_ids, full_docnames)
        cs = sorted(cs, key=lambda x: x[1], reverse=True)
        cs = [{'query': query, 'context': q['content'], 'score': s, 'id': i, 'docname': d} for q, s, i, d in cs]
        return cs




from torch.utils.data import Dataset
from typing import NamedTuple
from retrieval.utils import Encoding, remove_padding, collate
import logging
logging.basicConfig(format='%(levelname)s :: %(asctime)s :: %(message)s', level=logging.WARNING)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class InferenceQuery(NamedTuple):
    query: str
    context: str
    id: str
    docname: str

class InferenceDataset(Dataset):
    def __init__(self, query, contexts, tokenizer, max_length=512):
        self.load_samples(query, contexts)
        self.max_length = max_length
        self.tokenizer = tokenizer

    def load_samples(self, query, contexts):
        self.samples = []
        for c in contexts:
            self.samples.append(InferenceQuery(query, c['content'], c['id'], c['pdf_name']))

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        sample = self.samples[idx]
        query, context, id, docname = sample
        x = self.tokenizer.encode_plus(
            query, context, return_tensors="pt", max_length=self.max_length, truncation=True, padding=True
        )
        x = remove_padding(x, self.tokenizer.pad_token_id)
        return x, id, docname

    @classmethod
    def collate(cls, samples):
        xs = [x[0] for x in samples]
        ids = [x[1] for x in samples]
        docnames = [x[2] for x in samples]
        padding_value = 0 # should probably replace this with a partial application in the collate call, but whatever for now
        return collate(xs, padding_value), ids, docnames




"""
Retriever interface
"""

class Retriever:
    def search(self, query):
        raise NotImplementedError('Use a subclass of Retriever')

    def rerank(self, query, contexts):
        raise NotImplementedError('Use a subclass of Retriever')

    def build_index(self, input_path):
        raise NotImplementedError('Use a subclass of Retriever')

    def delete(self, dataset_id):
        raise NotImplementedError('Use a subclass of Retriever')




from retrieve import Retrieval
import os

if __name__ == "__main__":
    dataset_id = os.environ['DATASET_ID']
    doc_idx_path = '/index_dir/' + dataset_id + '/documents_jsonl/lucene-index'
    context_idx_pth = '/index_dir/' + dataset_id + '/contexts_jsonl/lucene-index'
    doc2odoc_pth = '/index_dir/' + dataset_id + '/id_to_pdf.pkl'
    ctx2octx_pth = '/index_dir/' + dataset_id + '/id_to_context.pkl'
    octx2odoc_pth = '/index_dir/'+ dataset_id + '/context_to_doc.pkl'
    retrieval = Retrieval(doc_idx_path, context_idx_pth, doc2odoc_pth, ctx2octx_pth, octx2odoc_pth, k1=1, k2=1)
    result = retrieval.search('explosive')
    for r in result:
        print(len(r))
        print(r[0], r[1])
    # Result is a list of list : [[actual_doc_id, actual_context_id, content, score], [actual_doc_id, actual_context_id, content, score], ....]
    #print(result)



"""
This retriever does not currently adhere to the Retriever interface
"""
from pyserini.search import pysearch

class PyseriniRetriever(Retriever):
    def __init__(self, doc_idx_path, context_idx_pth, doc2odoc_pth, ctx2octx_pth, octx2odoc_pth, k1=5, k2=100):
        self.doc_searcher = pysearch.SimpleSearcher(doc_idx_path)
        self.context_searcher = pysearch.SimpleSearcher(context_idx_pth)
        with open(doc2odoc_pth, 'rb') as handle:
            self.doc2odoc_map = pickle.load(handle)
        with open(ctx2octx_pth, 'rb') as handle:
            self.ctx2octx_map = pickle.load(handle)
        with open(octx2odoc_pth, 'rb') as handle:
            self.octx2odoc_map = pickle.load(handle)
        self.k1 = k1
        self.k2 = k2

    def search(self, query, start_index=0):
        k = 10 # Number of contexts to retrieve
        doc_hits = self.doc_searcher.search(query, k=self.k1)
        results = []
        doc_ids = []
        for i in range(0, min(len(doc_hits), self.k1)):
            index = doc_hits[i].docid
            doc_ids.append(int(self.doc2odoc_map[int(index)]))

        context_hits = self.context_searcher.search(query, k=self.k2)

        for i in range(start_index, min(len(context_hits), self.k2)):
            index = context_hits[i].docid
            content = context_hits[i].raw
            score = context_hits[i].score
            actual_context_id = self.ctx2octx_map[int(index)]
            actual_doc_id = self.octx2odoc_map[int(actual_context_id)]
            if actual_doc_id in doc_ids:
                results.append([actual_doc_id, actual_context_id, content, score, i])
        return results




from setuptools import setup, find_packages
setup(
    name="cosmos-retrieve",
    version="0.1",
    author="Ankur Goswami",
    author_email="ankur.goswami12@gmail.com",
    description="The COSMOS retrieval package",
    url="https://github.com/UW-COSMOS/Cosmos",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
